package org.expeditee.settings;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.gui.AttributeValuePair;
import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameUtils;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.items.widgets.Password;
import org.expeditee.reflection.PackageLoader;
import org.expeditee.setting.GenericSetting;
import org.expeditee.setting.Setting;
import org.expeditee.setting.VariableSetting;

public abstract class Settings {

	public static final String SETTINGS_PACKAGE_PARENT = "org.expeditee.";
	public static final String SETTINGS_PACKAGE = SETTINGS_PACKAGE_PARENT + "settings.";
	
	private static final class PageDescriptor {
		
		private final HashMap<String, Setting> settings = new HashMap<String, Setting>();
		private final List<String> orderedEntries = new LinkedList<String>();
		private final List<VariableSetting> hasDefault = new LinkedList<VariableSetting>();
		private final Method onParsed;
		
		private PageDescriptor(Class<?> clazz) {
			
			// populate map of settings
			for(Field f : clazz.getFields()) {
				// only allow valid types
				if(! Setting.class.isAssignableFrom(f.getType())) {
					continue;
				}
				try {
					Setting s = (Setting) f.get(null);
	                settings.put(f.getName().toLowerCase(), s);
	                if(s instanceof VariableSetting && ((VariableSetting) s).hasDefault()) {
	                	hasDefault.add((VariableSetting) s);
	                }
	                orderedEntries.add(f.getName());
                } catch (Exception e) {
	                e.printStackTrace();
                }
			}
			Method m = null;
			try {
				m = clazz.getMethod("onParsed", Text.class);
			} catch(Exception e) {
				// System.err.println(clazz.getName() + " has no onParsed(Text t) callback");
			}
			this.onParsed = m;
		}
	}
	private static HashMap<String, PageDescriptor> _pages = new HashMap<String, PageDescriptor>();

	public static void Init() {
		try {
			for(Class<?> clazz : PackageLoader.getClassesNew(SETTINGS_PACKAGE)) {
				// Ignore this class since it's the controller
				if(clazz.equals(Settings.class)) {
					continue;
				}
				String settingsPage = clazz.getPackage().getName().toLowerCase().substring(SETTINGS_PACKAGE_PARENT.length());
				// System.out.println(settingsPage + " : " + clazz.getName());
				_pages.put(settingsPage, new PageDescriptor(clazz));
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
		FrameUtils.ParseProfile(FrameIO.LoadProfile(UserSettings.ProfileName.get()));
	}
	
	public static void parseSettings(Text text) {
		parseSettings(text, "");
	}
	
	/**
	 * Sets all the simple settings.
	 * 
	 * @param text
	 * @param prefix
	 */
	private static void parseSettings(Text text, String prefix) {
		
		Frame child = text.getChild();
		if(child == null) {
			return;
		}
		String settingsPage = prefix + text.getText().trim().toLowerCase().replaceAll("^@", "");
		PageDescriptor pd = _pages.get(settingsPage);
		if(pd == null) {
			return;
		}
		try {
			if (text.getChild() == null) {
				return;
			}
			List<VariableSetting> toDefault = new LinkedList<VariableSetting>(pd.hasDefault);
			// set the fields
			for (Text t : text.getChild().getBodyTextItems(true)) {
				AttributeValuePair avp = new AttributeValuePair(t.getText(), false);
				try {
					// System.out.println(avp.getAttributeOrValue().trim().toLowerCase().replaceAll("^@", ""));
					Setting s = pd.settings.get(avp.getAttributeOrValue().trim().toLowerCase().replaceAll("^@", ""));
					if(s == null) {
						continue;
					}
					if(s.setSetting(t)) {
						toDefault.remove(s);
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			// set any unset fields to their default
			for(VariableSetting s : toDefault) {
				try {
					// System.out.println("Resetting " + s.getTooltip());
		            s.reset();
		            // System.out.println("Set " + f.getName() + " to default value " + f.get(null));
	            } catch (Exception e) {
		            e.printStackTrace();
	            }
			}
			// call the onParsed method if one exists
			if(pd.onParsed != null) {
				pd.onParsed.invoke(null, text);
			}
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
		// if the page was a settings page, check if it has any subpages
		for(Text t : child.getTextItems()) {
			parseSettings(t, settingsPage + ".");
		}
	}
	
	public static Text generateSettingsTree() {
		return generateSettingsTree("settings", DisplayIO.getCurrentFrame().createNewText("Settings"));
	}
	
	/**
	 * 
	 * Generates settings tree
	 * TODO: merge with @see FrameCreator
	 * 
	 */
	private static Text generateSettingsTree(String page, Text text) {
		Frame frame = FrameIO.CreateNewFrame(text);
		text.setLink(frame.getName());
		
		int x = 50, y = 100;
		int dY = UserSettings.ItemTemplate.get().getPolygon().getBounds().height + 1;
		int maxY = UserSettings.InitialHeight.get() - 150;
		
		// add subpages of the current page
		for(String k : _pages.keySet()) {
			if(k.startsWith(page.toLowerCase()) && !k.equals(page)) {
				String name = k.substring(page.length() + 1);
				if(name.indexOf('.') != -1) {
					continue;
				}
				System.out.println("Generating " + name);
				generateSettingsTree(k, frame.addText(x, y, name.substring(0, 1).toUpperCase() + name.substring(1), null));
				y += dY;
				if(y >= maxY) {
					x += 300;
					y = 100;
				}
			}
		}
		
		x += 200;
		y = 100;
		
		// add settings of the current page
		PageDescriptor pd = _pages.get(page);
		if(pd != null) {
			for(String str : pd.orderedEntries) {
				String key = str.toLowerCase();
				Setting s = pd.settings.get(key);
				if(s == null) {
					continue;
				}
				String name = str.substring(0, 1).toUpperCase() + str.substring(1);
				String value = "";
				if(s instanceof GenericSetting && ((GenericSetting) s).isPrimitive()) {
					System.out.println("VALUE: " + ((GenericSetting) s).get());
					if(((GenericSetting) s).get() != null) {
						value = ": " + ((GenericSetting) s).get();
					} else {
						value = ": ";
					}
				}
				if(key.equals("pass")) {
					Text t = frame.addText(x, y, "iw: org.expeditee.items.widgets.Password", null);
					Password pw = new Password(t, null);
					pw.setPassword((String) value);
					frame.removeItem(t);
					frame.addAllItems(pw.getItems());
					y += pw.getHeight();
				} else if(s instanceof GenericSetting && ((GenericSetting) s).getType().equals(Text.class)) {
					Text t = ((GenericSetting<Text>) s).get();
					if(t == null) continue;
					t = t.copy();
					int h = t.getPolygon().getBounds().height + 1;
					t.setID(frame.getNextItemID());
					t.setX(x);
					t.setY(y + h / 2);
					t.setText(name);
					frame.addItem(t);
					y += h;
				} else {
					frame.addText(x, y + dY / 2, name + value, null);
					y += dY;
				}
				if(y >= maxY) {
					x += 300;
					y = 100;
				}
			}
		}
		
		FrameIO.SaveFrame(frame);
		return text;
	}
}
